import 'dart:ffi';
import 'dart:io';
import 'dart:math';

import 'package:flutter_player/ffmpeg.dart';
import 'package:torrent/torrent.dart';
import 'package:logging/logging.dart';

class TorrentRequest extends IsolateProtocolRequest {
  BtTorrent torrent;
  int offset = 0;
  bool closed = false;
  RandomAccessFile file;

  static final agent = BtAgentCore();

  TorrentRequest._new(int bufferSize) : super(bufferSize);

  static Future<TorrentRequest> open(
      String url, void onUpdate(BtTorrent torrent)) async {
    BtLog.setLevel(Level.FINE);
    final data = await File(url).readAsBytes();
    final btInfo = BtParser().parse(data);
    final torrent = agent.addTorrent(btInfo);
    final req = TorrentRequest._new(
      torrent.info.metaInfo.pieceLengthInBytes,
    )..torrent = torrent;
    (() async {
      await req.torrent.verifyLocalData((_) => req.closed);
      req.torrent.startAllTrackers();
      var lastCount = -1;
      while (!req.closed) {
        if (lastCount != torrent.bitfield.trueCount) onUpdate(torrent);
        lastCount = torrent.bitfield.trueCount;
        await Future.delayed(Duration(milliseconds: 500));
        req.torrent.updateTimeout();
        req.torrent.updateRequests();
      }
      req.torrent.stopAllTrackers();
      torrent.storage.close();
    })();
    return req;
  }

  @override
  Future<int> read(Pointer<Uint8> buf, int size) async {
    if (offset >= torrent.size) return -1;
    final pieceIndex = offset ~/ torrent.pieceSize;
    final readSize = min(
        pieceIndex * torrent.pieceSize +
            torrent.getPieceSize(pieceIndex) -
            offset,
        size);
    if (readSize == 0) return -1;
    while (!closed && !torrent.bitfield.get(pieceIndex)) {
      await Future.delayed(Duration.zero);
      if (isCancel) return -11;
      torrent.download(pieceIndex,
          min(torrent.pieceCount - 1, (offset + size) ~/ torrent.pieceSize));
    }
    final read = await torrent.storage.read(readSize, offset);
    buf.asTypedList(read.length).setAll(0, read);
    offset += read.length;
    return read.length;
  }

  @override
  Future<int> seek(int offset, int whence) async {
    switch (whence) {
      case AVSEEK_SIZE:
        return torrent.size;
      default:
        this.offset = offset;
        return 0;
    }
  }

  @override
  Future<void> closeImpl() async {
    closed = true;
  }
}
