part of '../ffmpeg.dart';

extension ListFirstWhere<T> on Iterable<T> {
  T? firstWhereOrNull(bool Function(T) test) {
    try {
      return firstWhere(test);
    } on StateError {
      return null;
    }
  }
}

class FfmpegStream implements _IsolateEncodable {
  _CodecContext? __codec;
  final int index;
  Pointer<AVStream>? _p;
  FfmpegStream._new(this.index, this._p);
  _CodecContext get _codec {
    if (_p == null) throw Exception("StreamInfo destroyed");
    if (__codec == null) __codec = _CodecContext._new(_p!.codecpar);
    return __codec!;
  }

  int _getFramePts(_Frame frame) {
    return getFrameTimeStamp(frame._value, _p!);
  }

  int get codecType => _p!.codecpar.codec_type;

  @override
  Map _encode() => {
        #streamIndex: index,
        #streamCodecType: codecType,
      };

  void _close() {
    if (__codec != null) __codec!._close();
    __codec = null;
  }
}

class _AudioCodecContext extends _CodecContext {
  Pointer<Pointer<SwrContext>>? _swrCtx;
  int _srcChannelLayout = -1;
  int _srcFormat = -1;
  int _srcSampleRate = -1;
  _AudioClient? _audio;

  Pointer<Pointer<Uint8>>? _buffer;
  Pointer<Pointer<Uint8>>? _buffer1;
  Pointer<Uint32>? _bufferLen;
  Pointer<Uint32>? _bufferLen1;

  @override
  Future<PlayFrame> _playFrame(PlayFrame frame, Future? delay) {
    final srcChannelLayout = frame._frame._value.channel_layout;
    final srcFormat = frame._frame._value.format;
    final srcSampleRate = frame._frame._value.sample_rate;
    if (_audio == null) {
      _audio = _AudioClient();
      _audio!.start();
    }
    if (_swrCtx == null) {
      _swrCtx = malloc();
      _swrCtx!.value = Pointer.fromAddress(0);
    }
    if (_swrCtx!.value.address == 0 ||
        _srcChannelLayout != srcChannelLayout ||
        _srcFormat != srcFormat ||
        _srcSampleRate != srcSampleRate) {
      if (_swrCtx!.value.address != 0) swr_free(_swrCtx!);
      _swrCtx!.value = swr_alloc_set_opts(
          Pointer.fromAddress(0),
          av_get_default_channel_layout(_audio!.channels),
          _audio!.format,
          _audio!.sampleRate,
          frame._frame._value.channel_layout,
          frame._frame._value.format,
          frame._frame._value.sample_rate,
          0,
          Pointer.fromAddress(0));
      if (_swrCtx!.value.address == 0 || swr_init(_swrCtx!.value) < 0)
        throw Exception("cannot create SwrContext");
      _srcChannelLayout = srcChannelLayout;
      _srcFormat = srcFormat;
      _srcSampleRate = srcSampleRate;
    }
    if (_buffer == null || _buffer!.value.address == 0) {
      _buffer = malloc();
      _buffer!.value = Pointer.fromAddress(0);
    }
    if (_bufferLen == null || _buffer!.address == 0) {
      _bufferLen = malloc();
      _bufferLen!.value = 0;
    }
    final inp = frame._frame._value.extended_data;
    final inCount = frame._frame._value.nb_samples;
    final outCount = inCount * _audio!.sampleRate ~/ _srcSampleRate + 256;
    final outSize = av_samples_get_buffer_size(
        Pointer.fromAddress(0), _audio!.channels, outCount, _audio!.format, 0);
    if (outSize < 0) throw Exception("av_samples_get_buffer_size failed");
    av_fast_malloc(_buffer!, _bufferLen!, outSize);
    if (_buffer!.value.address == 0) throw Exception("av_fast_malloc failed");
    final nbSamples =
        swr_convert(_swrCtx!.value, _buffer!, outCount, inp, inCount);
    // swap buffer -> buffer1
    final buffer = _buffer;
    final bufferLen = _bufferLen;
    _buffer = _buffer1;
    _bufferLen = _bufferLen1;
    _buffer1 = buffer;
    _bufferLen1 = bufferLen;
    return Future.value(delay).then((_) async {
      await _audio!.writeBuffer(
        buffer!.value,
        nbSamples,
        av_get_bytes_per_sample(_audio!.format) * _audio!.channels,
      );
      final timeStamp = frame.timeStamp -
          _audio!.getCurrentPadding() * AV_TIME_BASE ~/ _audio!.sampleRate;
      return PlayFrame._new(timeStamp < 0 ? 0 : timeStamp, frame._frame, null);
    });
  }

  @override
  void _close() {
    super._close();
    if (_audio != null) {
      _audio!.stop();
      _audio!.close();
      _audio = null;
    }
    if (_buffer != null) {
      if (_buffer!.value.address != 0) av_free(_buffer!.value);
      malloc.free(_buffer!);
    }
    _buffer = null;
    if (_buffer1 != null) {
      if (_buffer1!.value.address != 0) av_free(_buffer1!.value);
      malloc.free(_buffer1!);
    }
    _buffer1 = null;
    if (_bufferLen != null) malloc.free(_bufferLen!);
    _bufferLen = null;
    if (_bufferLen1 != null) malloc.free(_bufferLen1!);
    _bufferLen1 = null;
    if (_swrCtx != null) {
      if (_swrCtx?.value.address != null) swr_free(_swrCtx!);
      malloc.free(_swrCtx!);
      _swrCtx = null;
    }
  }
}

class _VideoCodecContext extends _CodecContext {
  _Frame? __frame;

  _Frame get _frame {
    if (_ctx == null) throw Exception("CodecContext destroyed");
    if (__frame == null) {
      _freeSwsCtx();
      __frame = _Frame._new();
      __frame!._fmt = AV_PIX_FMT_RGBA;
      __frame!._width = _ctx!.value.width;
      __frame!._height = _ctx!.value.height;
      final bufSize = av_image_get_buffer_size(
          AV_PIX_FMT_RGBA, __frame!._width!, __frame!._height!, 1);
      __frame!._buffer = malloc.allocate<Uint8>(bufSize);
      av_image_fill_arrays(
          __frame!._value.data, // dst data[]
          __frame!._value.linesize, // dst linesize[]
          __frame!._buffer!, // src buffer
          AV_PIX_FMT_RGBA, // pixel format
          __frame!._width!, // width
          __frame!._height!, // height
          1 // align
          );
    }
    return __frame!;
  }

  Pointer<SwsContext>? _swsCtx;
  void _freeSwsCtx() {
    if (_swsCtx != null && _swsCtx!.address != 0) {
      sws_freeContext(_swsCtx!);
    }
    _swsCtx = null;
  }

  @override
  void _close() {
    super._close();
    if (__frame != null) __frame!.close();
    __frame = null;
    _freeSwsCtx();
  }

  @override
  Future<PlayFrame?> _playFrame(PlayFrame frame, Future? delay) async {
    if (_ctx == null || _ctx!.address == 0) return null;
    if (_swsCtx == null || _swsCtx!.address == 0) {
      _swsCtx = sws_getContext(
          _ctx!.value.width, // src width
          _ctx!.value.height, // src height
          _ctx!.value.pix_fmt, // src format
          _frame._width!, // dst width
          _frame._height!, // dst height
          _frame._fmt!, // dst format
          SWS_POINT, // flags
          Pointer.fromAddress(0), // src filter
          Pointer.fromAddress(0), // dst filter
          Pointer.fromAddress(0) // param
          );
    }
    if (_swsCtx == null || _swsCtx!.address == 0)
      throw Exception("cannot create SwsContext");
    await delay;
    sws_scale(
        _swsCtx!, // sws context
        frame._frame._value.data, // src slice
        frame._frame._value.linesize, // src stride
        0, // src slice y
        _frame._height!, // src slice height
        _frame._value.data, // dst planes
        _frame._value.linesize // dst strides
        );
    return PlayFrame._new(
      frame.timeStamp,
      _frame,
      null,
    );
  }
}

abstract class _CodecContext {
  Pointer<Pointer<AVCodecContext>>? _ctx;

  static _new(Pointer<AVCodecParameters> codecpar) {
    final pCodec = avcodec_find_decoder(codecpar.codec_id);
    if (pCodec.address == 0) throw Exception("avcodec_find_decoder failed");
    final pCodecCtx = malloc<Pointer<AVCodecContext>>();
    pCodecCtx.value = avcodec_alloc_context3(pCodec);
    var ret = avcodec_parameters_to_context(pCodecCtx.value, codecpar);
    if (ret != 0) throw Exception("avcodec_parameters_to_context failed: $ret");
    ret = avcodec_open2(pCodecCtx.value, pCodec, Pointer.fromAddress(0));
    if (ret != 0) throw Exception("avcodec_open2 failed: $ret");
    switch (codecpar.codec_type) {
      case AVMediaType.AVMEDIA_TYPE_AUDIO:
        return _AudioCodecContext().._ctx = pCodecCtx;
      case AVMediaType.AVMEDIA_TYPE_VIDEO:
        return _VideoCodecContext().._ctx = pCodecCtx;
    }
  }

  Future<PlayFrame?> _playFrame(PlayFrame frame, Future? delay);

  void _close() {
    if (_ctx == null) return;
    avcodec_free_context(_ctx!);
    malloc.free(_ctx!);
    _ctx = null;
  }
}

class _Frame {
  int? _fmt;
  Pointer<Pointer<AVFrame>>? _p;
  Pointer<AVFrame> get _value => _p!.value;
  Pointer<Uint8>? _buffer;
  int? _width;
  int? _height;
  _Frame._new() {
    _p = malloc<Pointer<AVFrame>>();
    _p!.value = av_frame_alloc();
  }

  void close() {
    if (_p == null) return;
    if (_buffer != null) malloc.free(_buffer!);
    _buffer = null;
    av_frame_free(_p!);
    malloc.free(_p!);
    _p = null;
  }
}

class PlayFrame extends _IsolateEncodable {
  final int timeStamp;
  final _Frame _frame;
  final _CodecContext? _codec;
  PlayFrame._new(this.timeStamp, this._frame, this._codec);

  get buffer => _frame._buffer;
  get width => _frame._width;
  get height => _frame._height;

  @override
  Map _encode() => {
        #frameWidth: width,
        #frameHeight: height,
        #frameBuffer: buffer?.address,
        #frameTimeStamp: timeStamp,
      };
}

class _PTS {
  int _relate = 0;
  int _absolute = DateTime.now().millisecondsSinceEpoch;
  bool playing = false;
  bool _skipAudio = false;
  final List<FfmpegStream> streams;
  _PTS(this.streams, [this._relate = 0]);

  void update(int relate) {
    _relate = relate;
    _absolute = DateTime.now().millisecondsSinceEpoch;
  }

  int ptsNow() =>
      (DateTime.now().millisecondsSinceEpoch - _absolute) *
          AV_TIME_BASE ~/
          1000 +
      _relate;
}

abstract class ProtocolRequest {
  final int bufferSize;
  int read(Pointer<Uint8> buf, int size);
  int seek(int offset, int whence);
  Future close();
  ProtocolRequest(
    this.bufferSize,
  );
}

final Map<int, ProtocolRequest> _protocolRequests = {};
const _readFailReturn = -1;
int _ioReadPacket(int opaque, Pointer<Uint8> buf, int bufSize) {
  return _protocolRequests[opaque]?.read(buf, bufSize) ?? _readFailReturn;
}

const _seekFailReturn = -1;
int _ioSeek(int opaque, int offset, int whence) {
  return _protocolRequests[opaque]?.seek(offset, whence) ?? _seekFailReturn;
}

class FormatContext {
  final String url;
  Pointer<Pointer<AVFormatContext>>? _ctx;
  Pointer<AVIOContext>? _ioCtx;
  ProtocolRequest? _req;
  final void Function(PlayFrame)? ___onFrame;
  _IsolateFunction? __onFrame;
  _IsolateFunction get _onFrame {
    if (__onFrame == null && ___onFrame != null)
      __onFrame = _IsolateFunction._new(___onFrame);
    return __onFrame!;
  }

  List<FfmpegStream>? _streams;

  FormatContext(this.url,
      {void Function(PlayFrame)? onFrame, ProtocolRequest? request})
      : ___onFrame = onFrame {
    _ctx = malloc<Pointer<AVFormatContext>>();
    _ctx!.value = Pointer.fromAddress(0);
    if (request != null) {
      _req = request;
      final idReq = identityHashCode(request);
      final bufferSize = request.bufferSize;
      if (bufferSize <= 0) throw Exception('buffer size must larger than 0');
      _protocolRequests[idReq] = request;
      Pointer<Uint8> _buffer = av_malloc(bufferSize).cast();
      _ioCtx = avio_alloc_context(
          _buffer,
          bufferSize,
          0,
          idReq,
          Pointer.fromFunction(_ioReadPacket, _readFailReturn),
          Pointer.fromAddress(0),
          Pointer.fromFunction(_ioSeek, _seekFailReturn));
    }
    _ctx!.value = avformat_alloc_context();
    if (_ioCtx != null) _ctx!.value.pb = _ioCtx;

    final _url = url.toNativeUtf8();
    int ret = avformat_open_input(
        _ctx!, _url, Pointer.fromAddress(0), Pointer.fromAddress(0));
    malloc.free(_url);
    if (ret != 0) throw Exception("avformat_open_input failed: $ret");
  }

  void _checkCtx() {
    if (_ctx == null || _ctx!.value.address == 0)
      throw Exception("context not initialized");
  }

  _PTS? _pts;
  Future? _playingFuture;

  bool isPlaying() => _pts?.playing ?? false;
  int getPosition() => _pts?.ptsNow() ?? 0;

  int duration() => _ctx?.value.duration ?? 0;

  Future play(
    List<FfmpegStream> streams,
  ) async {
    await pause();
    _pts = _PTS(streams, 0);
    _seekTo(0);
    _playingFuture = resume();
    return _playingFuture;
  }

  Future pause() {
    _pts?.playing = false;
    return Future.value(_playingFuture);
  }

  Future seekTo(
    int ts, {
    int streamIndex = -1,
    int minTs = INT64_MIN,
    int maxTs = INT64_MAX,
    int flags = 0,
  }) async {
    if (_pts == null) throw Exception("no pts data");
    final pauseAtSeekTo = _pts!.playing;
    final pts = _PTS(_pts!.streams, ts);
    _pts = pts;
    for (var stream in pts.streams) {
      final codec = stream.__codec?._ctx?.value;
      if (codec != null) avcodec_flush_buffers(codec);
    }
    _seekTo(
      ts,
      streamIndex: streamIndex,
      minTs: minTs,
      maxTs: maxTs,
      flags: flags,
    );
    final _pauseAtStreamIndex = pts.streams
            .firstWhereOrNull((e) =>
                (streamIndex == -1 || e.index == streamIndex) &&
                e._p?.codecpar.codec_type == AVMediaType.AVMEDIA_TYPE_VIDEO)
            ?.index ??
        -1;
    if (_pauseAtStreamIndex != -1) {
      pts._skipAudio = true;
      Completer ret = Completer();
      _playingFuture = _resume((id) {
        if ((id == null || id == _pauseAtStreamIndex) && !ret.isCompleted) {
          pts._skipAudio = false;
          if (!pauseAtSeekTo) {
            pts.playing = false;
          }
          ret.complete();
        }
      });
      return ret.future;
    } else if (pauseAtSeekTo) {
      _playingFuture = resume();
    }
  }

  void _seekTo(
    int ts, {
    int streamIndex = -1,
    int minTs = INT64_MIN,
    int maxTs = INT64_MAX,
    int flags = 0,
  }) async {
    avformat_seek_file(
      _ctx!.value,
      streamIndex,
      minTs,
      ts,
      maxTs,
      flags,
    );
  }

  Future resume() => _resume(null);

  Future _resume(void onNextFrame(int)?) async {
    final pts = _pts!;
    final _isPlaying = () {
      return (this._pts == pts) && pts.playing;
    };
    final packet = av_packet_alloc();
    var frame = _Frame._new();
    try {
      Map<FfmpegStream, Future Function(PlayFrame)> streamUpdators = {};
      pts.streams.forEach((stream) {
        Future? lastUpdate;
        streamUpdators[stream] = (PlayFrame frame) {
          final _lastUpdate = lastUpdate;
          final ptsNow = pts.ptsNow();
          final timeStamp = frame.timeStamp;
          final comp = () => frame._codec!._playFrame(frame, _lastUpdate);
          final codecType = stream._p!.codecpar.codec_type;
          lastUpdate = codecType == AVMediaType.AVMEDIA_TYPE_VIDEO &&
                  timeStamp > ptsNow
              ? Future.delayed(
                  Duration(
                      milliseconds:
                          (timeStamp - ptsNow) * 1000 ~/ AV_TIME_BASE),
                  () => _isPlaying() ? comp() : null)
              : codecType == AVMediaType.AVMEDIA_TYPE_AUDIO && pts._skipAudio
                  ? Future.value()
                  : comp();
          lastUpdate!.then((_frame) {
            if (_isPlaying()) {
              if (__onFrame != null && _frame is PlayFrame) _onFrame(_frame);
              if (_frame is PlayFrame &&
                  stream._p!.codecpar.codec_type ==
                      AVMediaType.AVMEDIA_TYPE_AUDIO)
                pts.update(_frame.timeStamp);
              if (onNextFrame != null) onNextFrame(stream.index);
            }
            frame._frame.close();
          });
          return Future.value(_lastUpdate);
        };
      });
      pts.playing = true;
      while (_isPlaying() &&
          _ctx != null &&
          av_read_frame(_ctx!.value, packet) == 0) {
        final streamIndex = packet.stream_index;
        final stream =
            pts.streams.firstWhereOrNull((s) => s.index == streamIndex);
        if (stream != null) {
          var ret = avcodec_send_packet(stream._codec._ctx!.value, packet);
          if (ret != 0) throw Exception("avcodec_send_packet failed: $ret");
          if (0 ==
              avcodec_receive_frame(
                stream._codec._ctx!.value,
                frame._value,
              )) {
            final timeStamp = stream._getFramePts(frame);
            await streamUpdators[stream]!
                .call(PlayFrame._new(timeStamp, frame, stream._codec));
            frame = _Frame._new();
          }
        }
        av_packet_unref(packet);
        await Future.delayed(Duration.zero);
      }
    } finally {
      __onFrame?.call(null);
      onNextFrame?.call(null);
      pts.playing = false;
      frame.close();
      av_free(packet);
    }
  }

  List<FfmpegStream> getStreams() {
    _checkCtx();
    if (_streams == null) {
      final ret =
          avformat_find_stream_info(_ctx!.value, Pointer.fromAddress(0));
      if (ret != 0) throw Exception("avformat_find_stream_info failed: $ret");
      _streams = [];
      final nbStreams = _ctx!.value.nb_streams;
      final streams = _ctx!.value.streams;
      for (var i = 0; i < nbStreams; ++i) {
        final stream = streams.elementAt(i).value;
        _streams!.add(FfmpegStream._new(i, stream));
      }
    }
    return _streams!;
  }

  close() async {
    await pause();
    if (_ctx == null || _ctx!.value.address == 0) return;
    if (_ioCtx != null && _ioCtx!.address != 0) av_free(_ioCtx!);
    _streams!.forEach((s) => s._close());
    avformat_close_input(_ctx!);
    malloc.free(_ctx!);
    _ctx = null;
    __onFrame?.destroy();
    __onFrame = null;
    await _req?.close();
    _protocolRequests.remove(identityHashCode(_req));
    _req = null;
  }
}
