import 'dart:convert';
import 'dart:ffi';
import 'dart:io';
import 'package:flutter_player/ffmpeg.dart';
import 'package:flutter_test/flutter_test.dart';

void main() {
  test('make', () async {
    final utf8Encoding = Encoding.getByName('utf-8');
    var cmakePath = 'cmake';
    if (Platform.isWindows) {
      var vsDir = Directory('C:/Program Files (x86)/Microsoft Visual Studio/');
      vsDir = (vsDir.listSync().firstWhere((e) => e is Directory) as Directory)
          .listSync()
          .last as Directory;
      cmakePath = vsDir.path +
          '/Common7/IDE/CommonExtensions/Microsoft/CMake/CMake/bin/cmake.exe';
    }
    final buildDir = './build';
    var result = Process.runSync(
      cmakePath,
      ['-S', './', '-B', buildDir],
      workingDirectory: 'test',
      stdoutEncoding: utf8Encoding,
      stderrEncoding: utf8Encoding,
    );
    stdout.write(result.stdout);
    stderr.write(result.stderr);
    expect(result.exitCode, 0);

    result = Process.runSync(
      cmakePath,
      ['--build', buildDir, '--verbose'],
      workingDirectory: 'test',
      stdoutEncoding: utf8Encoding,
      stderrEncoding: utf8Encoding,
    );
    stdout.write(result.stdout);
    stderr.write(result.stderr);
    expect(result.exitCode, 0);
  });
  test('changeDirectory', () async {
    Directory.current = './test/build/Debug';
  });
  test('avformat', () async {
    final url = 'D:/CloudMusic/seven oops - オレンジ.flac';
    final request = await FileRequest.open(url);
    final ctx = IsolateFormatContext(url, request: request);
    final astream = (await ctx.getStreams()).firstWhere(
        (infos) => infos.codecType == AVMediaType.AVMEDIA_TYPE_AUDIO);
    ctx.play([astream]);
    await Future.delayed(Duration(seconds: 5));
    await ctx.seekTo(30 * AV_TIME_BASE);
    await Future.delayed(Duration(seconds: 5));
    await ctx.close();
  });
}

class FileRequest extends IsolateProtocolRequest {
  RandomAccessFile? file;

  FileRequest._new(this.file, [int bufferSize = 32768]) : super(bufferSize);

  static Future<FileRequest> open(String url) async {
    return FileRequest._new(await File(url).open());
  }

  @override
  Future<int> read(Pointer<Uint8> buf, int size) async {
    final ret = file!.readIntoSync(buf.asTypedList(size));
    if (ret == 0) return -1;
    return ret;
  }

  @override
  Future<int> seek(int offset, int whence) async {
    switch (whence) {
      case AVSEEK_SIZE:
        return file!.lengthSync();
      default:
        file!.setPositionSync(offset);
        return 0;
    }
  }

  @override
  Future<void> closeImpl() async {
    file?.close();
    file = null;
  }
}
